#!/usr/bin/python3
import random
import sys
import inspect

# Don't implement your own encryption!

# encrypt (or decrypt) a file with a one time pad that is determined by the
# python seed() function.  This is completely deterministic so it isn't
# useful for real encryption, but should be enough to avoid naive network
# inspection.

# The one time pad is infinitely long, determined by python's randomization
# function, so it shouldn't be succeptible to any attacks based on having
# a smaller pad than data source.  It is susceptible to somebody trying
# every possible random python seed, but we are offsetting that with
# some user configurable SALT

# change these salt values to something else.  This overlays a repeating
# pattern over the output of random.range() that should make it more difficult
# to reverse even if the method is known.
SALT = 42           # added to the byte value, no point going > 255
PEPPER = 13         # alters the SALT value by this amount every loop, no point going > 255
BURNOFF = 99        # burn off this many values before starting

# ref:  https://en.wikipedia.org/wiki/List_of_file_signatures

FILE_SIGNATURES = {'wk1': b'\x00\x00\x02\x00\x06\x04\x06\x00\x08\x00\x00\x00\x00\x00',
'wk3': b'\x00\x00\x1a\x00\x00\x10\x04\x00\x00\x00\x00\x00',
'wk4': b'\x00\x00\x1a\x00\x02\x10\x04\x00\x00\x00\x00\x00',
'wk5': b'\x00\x00\x1a\x00\x02\x10\x04\x00\x00\x00\x00\x00',
'123': b'\x00\x00\x1a\x00\x05\x10\x04',
'qxd': b'\x00\x00MMXPR',
'psafe3': b'PWS3',
'pcap': b'\xa1\xb2\xc3\xd4',
'pcapng': b'\n\r\r\n',
'rpm': b'\xed\xab\xee\xdb',
'sqlitedb': b'SQLite format 3\x00',
'bin': b'SP01',
'wad': b'IWAD',
'pic': b'\x00',
'pif': b'\x00',
'sea': b'\x00',
'ytr': b'\x00',
'pdb': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
'dba': b'\xbe\xba\xfe\xca',
'tda': b'\x00\x01DT',
'tdf': b'TDF$',
'tdef': b'TDEF',
'ico': b'\x00\x00\x01\x00',
'icns': b'icns',
'3gp': b'ftyp3g',
'3g2': b'ftyp3g',
'z': b'\x1f\x9d',
'lzh': b'-hl0-',
'bac': b'BACKMIKEDISK',
'idx': b'INDX',
'plist': b'bplist',
'bz2': b'BZh',
'gif': b'GIF87a',
'tif': b'MM\x00*',
'tiff': b'MM\x00*',
'cr2': b'II*\x00\x10\x00\x00\x00CR',
'cin': b'\x80*_\xd7',
'nui': b'NURUIMG',
'nup': b'NURUPAL',
'dpx': b'SDPX',
'exr': b'v/1\x01',
'bpg': b'BPG\xfb',
'jpg': b'\xff\xd8\xff\xdb',
'jpeg': b'\xff\xd8\xff\xdb',
'jp2': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'j2k': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'jpf': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'jpm': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'jpg2': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'j2c': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'jpc': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'jpx': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'mj2': b'\x00\x00\x00\x0cjP  \r\n\x87\n',
'lzip': b'LZIP',
'cpio': b'070707',
'exe': b'MZ',
'dll': b'MZ',
'mui': b'MZ',
'sys': b'MZ',
'scr': b'MZ',
'cpl': b'MZ',
'ocx': b'MZ',
'ax': b'MZ',
'iec': b'MZ',
'ime': b'MZ',
'rs': b'MZ',
'tsp': b'MZ',
'fon': b'MZ',
'efi': b'MZ',
'zip': b'PK\x03\x04',
'aar': b'PK\x03\x04',
'apk': b'PK\x03\x04',
'docx': b'PK\x03\x04',
'epub': b'PK\x03\x04',
'ipa': b'PK\x03\x04',
'jar': b'PK\x03\x04',
'kmz': b'PK\x03\x04',
'maff': b'PK\x03\x04',
'msix': b'PK\x03\x04',
'odp': b'PK\x03\x04',
'ods': b'PK\x03\x04',
'odt': b'PK\x03\x04',
'pk3': b'PK\x03\x04',
'pk4': b'PK\x03\x04',
'pptx': b'PK\x03\x04',
'usdz': b'PK\x03\x04',
'vsdx': b'PK\x03\x04',
'xlsx': b'PK\x03\x04',
'xpi': b'PK\x03\x04',
'rar': b'Rar!\x1a\x07\x01\x00',
'png': b'\x89PNG\r\n\x1a\n',
'com': b'\xc9',
'class': b'\xca\xfe\xba\xbe',
'txt': b'\xef\xbb\xbf',
'psafe3': b'%!PS',
'eps': b'%!PS-Adobe-3.0 EPSF-3.0',
'epsf': b'%!PS-Adobe-3.0 EPSF-3.0',
'chm': b'ITSF\x03\x00\x00\x00`\x00\x00\x00',
'hlp': b'?_',
'pdf': b'%PDF-',
'asf': b'0&\xb2u\x8ef\xcf\x11\xa6\xd9\x00\xaa\x00b\xcel',
'wma': b'0&\xb2u\x8ef\xcf\x11\xa6\xd9\x00\xaa\x00b\xcel',
'wmv': b'0&\xb2u\x8ef\xcf\x11\xa6\xd9\x00\xaa\x00b\xcel',
'ogg': b'OggS',
'oga': b'OggS',
'ogv': b'OggS',
'psd': b'8BPS',
'mp3': b'\xff\xfb',
'bmp': b'BM',
'dib': b'BM',
'iso': b'CD001',
'cdi': b'CD001',
'mgw': b'main.bs',
'nes': b'NES',
'd64': b'\xa02A\xa0\xa0\xa0',
'g64': b'GSR-1541',
'd81': b'\xa03D\xa0\xa0',
't64': b'C64 tape image file',
'crt': b'C64 CARTRIDGE   ',
'fits': b'SIMPLE  =                    T',
'flac': b'fLaC',
'mid': b'MThd',
'midi': b'MThd',
'doc': b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1',
'xls': b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1',
'ppt': b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1',
'msi': b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1',
'msg': b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1',
'dex': b'dex\n035\x00',
'vmdk': b'KDM',
'crx': b'Cr24',
'fh8': b'AGD3',
'cwk': b'\x06\x07\xe1\x00BOBO\x06\x07\xe1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01',
'toast': b'ER\x02\x00\x00\x00',
'xar': b'xar!',
'dat': b'PMOCCMOC',
'tar': b'ustar\x0000',
'tox3': b'tox3',
'mlv': b'MLVI',
'7z': b"7z\xbc\xaf'\x1c",
'gz': b'\x1f\x8b',
'xz': b'\xfd7zXZ\x00',
'lz4': b'\x04"M\x18',
'cab': b'MSCF',
'flif': b'FLIF',
'mkv': b'\x1aE\xdf\xa3',
'mka': b'\x1aE\xdf\xa3',
'mks': b'\x1aE\xdf\xa3',
'mk3d': b'\x1aE\xdf\xa3',
'webm': b'\x1aE\xdf\xa3',
'stg': b'MIL ',
'der': b'0\x82',
'crt': b'-----BEGIN CERTIFICATE-----',
'csr': b'-----BEGIN CERTIFICATE REQUEST-----',
'key': b'-----BEGIN PRIVATE KEY-----',
'ppk': b'PuTTY-User-Key-File-3:',
'pub': b'-----BEGIN SSH2 KEY-----',
'dcm': b'DICM',
'woff': b'wOFF',
'woff2': b'wOF2',
'wasm': b'\x00asm',
'lep': b'\xcf\x84\x01',
'swf': b'CWS',
'deb': b'!<arch>\n',
'rtf': b'{\\rtf1',
'ts': b'G',
'tsv': b'G',
'tsa': b'G',
'mpg': b'G',
'mpeg': b'G',
'm2p': b'\x00\x00\x01\xba',
'vob': b'\x00\x00\x01\xba',
'mp4': b'ftypisom',
'zlib': b'x\xda',
'lzfse': b'bvx2',
'orc': b'ORC',
'avro': b'Obj\x01',
'rc': b'SEQ6',
'p25': b'e\x87xV',
'obt': b'e\x87xV',
'pbt': b'xV4',
'pdt': b'xV4',
'pea': b'xV4',
'peb': b'xV4',
'pet': b'xV4',
'pgt': b'xV4',
'pict': b'xV4',
'pjt': b'xV4',
'pkt': b'xV4',
'pmt': b'xV4',
'ez2': b'EMX2',
'ez3': b'EMU3',
'luac': b'\x1bLua',
'alias': b'book\x00\x00\x00\x00mark\x00\x00\x00\x00',
'eml': b'Received:',
'tdef': b' \x02\x01b\xa0\x1e\xab\x07\x02\x00\x00\x00',
'kdb': b'7H\x03\x02\x00\x00\x00\x00X509KEY',
'zst': b'(\xb5/\xfd',
'rs': b'RSVKDATA',
'sml': b':)\n',
'srt': b'1\n00',
'vpk': b'4\x12\xaaU',
'ace': b'**ACE**',
'arj': b'`\xea',
'cab': b'ISc(',
'zoo': b'ZOO',
'pbm': b'P1\n',
'pgm': b'P2\n',
'wmf': b'\xd7\xcd\xc6\x9a',
'xcf': b'gimp xcf',
'xpm': b'/* XPM */',
'aff': b'AFF',
'Ex01': b'EVF2',
'e01': b'EVF',
'qcow': b'QFI',
'flv': b'FLV',
'vhd': b'connectix',
'vhdx': b'vhdxfile',
'isz': b'IsZ!',
'daa': b'DAA',
'evt': b'LfLe',
'evtx': b'ElfFile',
'sdb': b'sdbf',
'grp': b'PMCC',
'icm': b'KCMS',
'hiv': b'regf',
'pst': b'!BDN',
'drc': b'DRACO',
'grib': b'GRIB',
'grib2': b'GRIB',
'blend': b'BLENDER',
'jxl': b'\x00\x00\x00\x0cJXL \r\n\x87\n',
'ttf': b'\x00\x01\x00\x00\x00',
'tte': b'\x00\x01\x00\x00\x00',
'dfont': b'\x00\x01\x00\x00\x00',
'otf': b'OTTO',
'wim': b'MSWIM\x00\x00\x00\xd0\x00\x00\x00\x00',
'swm': b'MSWIM\x00\x00\x00\xd0\x00\x00\x00\x00',
'esd': b'MSWIM\x00\x00\x00\xd0\x00\x00\x00\x00',
'slob': b'!-1SLOB\x1f',
'voc': b'Creative Voice File\x1a\x1a\x00',
'au': b'.snd',
'snd': b'.snd',
'hazelrules': b'HZLR\x00\x00\x00\x18',
'flp': b'FLhd',
'flm': b'10LF',
'mny': b'\x00\x01\x00\x00MSISAM Database',
'accdb': b'\x00\x01\x00\x00Standard ACE DB',
'mdb': b'\x00\x01\x00\x00Standard Jet DB',
'drw': b'\x01\xff\x02\x04\x03\x02',
'dss': b'\x02dss',
'adx': b'\x03\x00\x00\x00APPR',
'indd': b'\x06\x06\xed\xf5\xd8\x1dF\xe5\xbd1\xef\xe7\xfet\xb7\x1d',
'mxf': b'\x06\x0e+4\x02\x05\x01\x01\r\x01\x02\x01\x01\x02',
'skf': b'\x07SKF',
'dtd': b'\x07dt2ddtd',
'wallet': b'\n\x16org.bitcoin.pr',
'nri': b'\x0eNeroISO',
'wks': b'\x0eWKS',
'sib': b'\x0fSIBELIUS',
'dsp': b'# Microsoft Developer Studio',
'amr': b'#!AMR',
'sil': b'#!SILK\n',
'hdr': b'#?RADIANCE\n',
'vbe': b'#@~^',
'cdb': b'\r\xf0\x1d\xc0',
'm3u': b'#EXTM3U',
'm3u8': b'#EXTM3U',
'm': b'mdf\x00',
'pak': b'KPKA',
'arc': b'ARC',
'pl': b'\xd0OPS',
}

def showhelp():
    print (f"""Syntax:

    required arguments:
    --mode ["encrypt" or "decrypt"]
    --file [filename]
    --seed [any string or int, python hashes this value]
    --suffix [for instance, "jpg", by default, "jpg"]


    optional encryption variables:
    --salt [int 0-255, default {SALT}]
    --pepper [int 0-255, default {PEPPER}]
    --burnoff [int, default {BURNOFF}]

    "donor" options:
    --donor [filename]  ( Use this file as byte/extension donor)
    --prefixlen X  ( specify the use of X bytes from the donor)

    
    "mini" code generators:
    --encryptor ( generates an encryptor )
    --decryptor ( generates a decryptor )

    "mini" syntax
    cat infile | mini.py [SEED] > outfile

    misc:
    --list (lists all supported suffixes)
    --help (you already found this)
    
""")
    

def decrypt(cipherbytes, seedval, prelen):
    global SALT
    global PEPPER
    global BURNOFF

    # strip off the header bytes, we know how big they are
    cipherbytes = cipherbytes[prelen:]

    # random.seed() results are shared across different systems, so it is a convenient OTP
    random.seed(seedval)

    # shift our OTP over by BURNOFF invocations
    for a in range(BURNOFF):
        x = random.randrange(256)

    outbuf = []

    for a in range(len(cipherbytes)):
        # SALT is added to each byte to hide the OTP.  PEPPER is used to adjust the amount SALT
        # every time it is used.  This should give us a max of 2^16 options although 
        SALT = (SALT + PEPPER) % 256
        rval = random.randrange(256)

        # subtract our pad value and salt in and take modulus
        b = (cipherbytes[a] - rval - SALT) % 256
        outbuf.append(b)
        
    outbuf = bytes(outbuf)
    return outbuf

def encrypt(plainbytes, seedval, prefix):
    global SALT
    global PEPPER
    global BURNOFF

    # random.seed() results are shared across different systems, so it is a convenient OTP
    random.seed(seedval)

    # shift our OTP over by BURNOFF invocations
    for a in range(BURNOFF):
        x = random.randrange(256)

    outbuf = []
    
    for a in range(len(plainbytes)):
        # SALT is added to each byte to hide the OTP.  PEPPER is used to adjust the amount SALT
        # every time it is used.  This should give us a max of 2^16 options although 
        SALT = (SALT + PEPPER) % 256
        rval = random.randrange(256)

        # add our pad value and salt in and take modulus
        b = (plainbytes[a] + rval + SALT) % 256

        outbuf.append(b)

    outbuf = bytes(outbuf)

    # add the prefix to the beginning of the buffer
    outbuf = prefix + outbuf
    
    return outbuf

#
# make a "miniature" encrypter/decrypter that works on stdin/stdout
#
def makemini(seedval, prefix, mode):
    global SALT
    global PEPPER
    global BURNOFF

    buffer = ""
    buffer += (f"""
import sys,random
SALT={SALT}
""")
    if mode == "encrypt":
        buffer += (inspect.getsource(encrypt))
        buffer += (f"""sys.stdout.buffer.write(encrypt(sys.stdin.buffer.read(), sys.argv[1], {prefix}))
""")
    else:
        buffer += (inspect.getsource(decrypt))
        buffer += (f"""sys.stdout.buffer.write(decrypt(sys.stdin.buffer.read(), sys.argv[1], {len(prefix)}))
""")
    
    # obfuscate our code a bit, also hardcode in some values
    buffer = buffer.replace("global PEPPER", "")
    buffer = buffer.replace("global BURNOFF", "")
    buffer = buffer.replace("SALT", "XXXXXXX")
    buffer = buffer.replace("PEPPER", str(PEPPER))
    buffer = buffer.replace("BURNOFF", str(BURNOFF))
    buffer = buffer.replace("len(prefix)", str(len(prefix)))
    buffer = buffer.replace("seedval", "XXXXXXXX")
    buffer = buffer.replace("cipherbytes", "XXXXXXXXX")
    buffer = buffer.replace("prefix", "XXXXXXXXXX")
    buffer = buffer.replace("outbuf", "XXXXXXXXXXX")
    buffer = buffer.replace("decrypt", "XXXXXXXXXXXX")
    buffer = buffer.replace("encrypt", "XXXXXXXXXXXXX")
    
    for line in buffer.split('\n'):
        line = line.rstrip()
        # get rid of empty lines and comments
        if len(line) > 1 and line.strip()[0] != "#":
            print(line)

def main(argv=None):

    global FILE_SIGNATURES
    global SALT
    global PEPPER
    global BURNOFF
    
    opt_mode = "encrypt"
    opt_filename = ""
    opt_seed = ""
    opt_suffix = "jpg"
    opt_makemini = ""
    opt_donorfile = ""
    opt_prefixlen = 1500   # one packet size.  I know this overshoots when you add tcp overhead
    
    for a in range(len(sys.argv)):
        if sys.argv[a] == "--mode":
            if sys.argv[a+1] in ("encrypt", "decrypt"):
                opt_mode = sys.argv[a+1]
            else:
                raise("Err, valid modes are encrypt and decrypt")
            a +=1
        if sys.argv[a] == "--file":
            opt_filename = sys.argv[a+1]
            a +=1
        if sys.argv[a] == "--salt":
            SALT = int(sys.argv[a+1])
            a +=1
        if sys.argv[a] == "--pepper":
            PEPPER = int(sys.argv[a+1])
            a +=1
        if sys.argv[a] == "--burnoff":
            BURNOFF = int(sys.argv[a+1])
            a +=1
        if sys.argv[a] == "--seed":
            opt_seed = sys.argv[a+1]
            a += 1
        if sys.argv[a] == "--suffix":
            opt_suffix = sys.argv[a+1]
            a += 1

        if sys.argv[a] == "--encryptor":
            opt_makemini = "encrypt"
        if sys.argv[a] == "--decryptor":
            opt_makemini = "decrypt"

        if sys.argv[a] == "--donor":
            opt_donorfile = sys.argv[a+1]
            a += 1

        if sys.argv[a] == "--prefixlen":
            opt_prefixlen = int(sys.argv[a+1])
            a += 1

        if sys.argv[a] == "--help":
            showhelp()
            sys.exit(0)

        if sys.argv[a] == "--list":
            k = list(FILE_SIGNATURES.keys())
            k.sort()
            print("Supported file extensions:")
            print (" ".join(k))
            sys.exit(0)
            
    opt_prefix = ""

    # so we can harddecode the prefix without having access to the donor file...
    if opt_prefixlen != 1500:
        opt_prefix = "x" * opt_prefixlen
        
    if opt_donorfile == "":
        opt_prefix = FILE_SIGNATURES[opt_suffix]
    else:
        with open(opt_donorfile, "rb") as f:
            opt_prefix = f.read(opt_prefixlen)
            opt_suffix = opt_donorfile.split(".")[-1]
        
    # if we want to make an encryptor or decryptor, do that and then quit
    if opt_makemini != "":
        makemini(opt_seed, opt_prefix, opt_makemini)
        sys.exit(0)
    
    ifile = open(opt_filename, "rb")
    data = ifile.read()
    ifile.close()


    filename = opt_filename
    if opt_mode == "encrypt":
        data = encrypt(data, opt_seed, opt_prefix)
        filename = filename.split('.')[0] + "." + opt_suffix

    if opt_mode == "decrypt":
        data = decrypt(data, opt_seed, len(opt_prefix))
        filename = filename.split('.')[0] + ".orig"  # original extension not preserved

    ofile = open(filename, "wb")
    ofile.write(data)
    ofile.close()
    print(f"Wrote {filename}")
    
if __name__ == '__main__':
        sys.exit(main(sys.argv))
